function [X,s,m,cost] = logm_iss_full(A,refine,short,rform)
%LOGM_ISS_FULL Matrix logarithm by inverse scaling and squaring method.
%   X = LOGM_ISS_FULL(A) computes the logarithm of A, for a matrix with no
%   nonpositive real eigenvalues, using the transformation-free form of
%   the inverse scaling and squaring method with Pade approximation.
%   Matrix square roots are computed by the product form of the
%   Denman-Beavers teration.
%   [X,S,M] = LOGM_ISS_FULL(A) returns the number S of square roots
%   computed and the degree M of the Pade approximant.
%   X = LOGM_ISS_FULL(A,REFINE,SHORT):
%       - REFINE: logical flag for use refined square root computation.
%       - SHORT:  logical flag for shorter code.

%   This code is intended for double precision.

%   Reference: A. H. Al-Mohy and N. J. Higham, Improved Inverse Scaling
%   and Squaring Algorithms for the Matrix Logarithm, MIMS EPrint 2011.83,
%   The University of Manchester, October 2011.
%   Name of corresponding algorithm in that paper:
%   Algorithm 5.2/iss_new.

%   Awad H. Al-Mohy and Nicholas J. Higham, October 19, 2011.
%   Bug in "short" case corrected 14-06-12.

xvals = [1.586970738772063e-005
         2.313807884242979e-003
         1.938179313533253e-002
         6.209171588994762e-002
         1.276404810806775e-001
         2.060962623452836e-001
         2.879093714241194e-001
         3.666532675959788e-001
         4.389227326152340e-001
         5.034050432047666e-001
         5.600071293013720e-001
         6.092525642521717e-001
         6.519202543720032e-001
         6.888477797186464e-001
         7.208340678820352e-001
         7.485977242539218e-001];

if nargin < 2 | isempty(refine), refine = 1; end
if nargin < 3 | isempty(short), short = 1; end
if nargin < 4 | isempty(rform), rform = 1; end

e = eig(A);
if any( imag(e) == 0 & real(e) <= 0 )
   error('A must not have any nonpositive real eigenvalues!')
end
mmax = 16;
n = length(A);

s = 0; its = 5; k = 0;
cost = 0; flag = 0;

d2 = normAm(A-eye(n),2)^(1/2); d3 = normAm(A-eye(n),3)^(1/3);
alpha2 = max(d2,d3);
if alpha2 <= xvals(2)
   m = find(alpha2 <= xvals(1:2),1); flag = 2;
   disp('m:'), m
   [norm(A-eye(n),1) d2 d3]
   A
end

while flag ~= 2

    if short

       % --------------------------------------
       % Shorter, but harder to understand code.
       d = zeros(6,1); alpha = zeros(5,1);
       flag = 0;
       if s > 0, d(3) = normAm(A-eye(n),3)^(1/3); end
       alpha(2) = inf; 
       for p=3:5
           d(p+1) = normAm(A-eye(n),p+1)^(1/(p+1));
           alpha(p) = max(d(p),d(p+1));
           eta = min(alpha(p-1),alpha(p));
           if eta <= xvals(16)
              kp = ceil( (p*(p-1)-1)/2 );
              j1 = find( eta <= xvals(kp:mmax),1 ) + kp-1;
              j2 = find( eta/2 <= xvals(kp:mmax),1 ) + kp-1;
              kp = ceil( (p*(p+1)-1)/2 );
              if 2*(j1-j2)/3 < its && j1 <= kp
                 m = j1; flag = 2; break;    % Break out of while loop.
              else
                 if 2*(j1-j2)/3 >= its && k < 2
                    k = k+1; flag = 1; break % Break out of for loop.
                 end
              end
           end
       end
       d6 = d(6); alpha5 = alpha(5);
       % --------------------------------------

    else

       % --------------------------------------
       % Longer, but easier to understand code.
       flag = 0;
       if s > 0, d3 = normAm(A-eye(n),3)^(1/3); end
       d4 = normAm(A-eye(n),4)^(1/4);
       alpha3 = max(d3,d4);
       if alpha3 <= xvals(mmax)
           j1 = find(alpha3 <= xvals(3:mmax),1) + 2;
           j2 = find(alpha3/2 <= xvals(3:mmax),1) + 2;
           if 2*(j1-j2)/3 < its && j1 <= 6
              m = j1; break;   % Break out of while loop.
           else
              if 2*(j1-j2)/3 >= its && k < 2
                 flag = 1; k = k+1;
              end
           end
       end
       if ~flag
           d5 = normAm(A-eye(n),5)^(1/5);
           alpha4 = max(d4,d5); eta4 = min(alpha3,alpha4);
           if eta4 <= xvals(mmax)
               j1 = find(eta4 <= xvals(6:mmax),1) + 5;
               j2 = find(eta4/2 <= xvals(6:mmax),1) + 5;
               if 2*(j1-j2)/3 < its && j1 <= 10
                  m = j1; break;
               else
                  if 2*(j1-j2)/3 >= its && k < 2
                     flag = 1; k = k+1;
                  end
               end
           end
       end
       if ~flag
           d6 = normAm(A-eye(n),6)^(1/6);
           alpha5 = max(d5,d6); eta5 = min(alpha4,alpha5);
           if eta5 <= xvals(mmax)
               j1 = find(eta5 <= xvals(10:mmax),1) + 9;
               j2 = find(eta5/2 <= xvals(10:mmax),1) + 9;
               if 2*(j1-j2)/3 < its && j1 <= 15
                  m = j1; break;
               else
                  if 2*(j1-j2)/3 >= its && k < 2
                     flag = 1; k = k+1;
                  end
               end
           end
       end
       % --------------------------------------
    end

    if ~flag
        d7 = normAm(A-eye(n),7)^(1/7);
        alpha6 = max(d6,d7); eta6 = min(alpha5,alpha6);
        if eta6 <= xvals(mmax)
            m = find(eta6 <= xvals(15:mmax),1) + 14;
        end
    end
    if flag ~= 2
       [A,~,its] = sqrtm_dbp(A,1); s = s+1;
       cost = cost + its;
       if refine
           if s == 1, Z0 = A - eye(n);   end
           if s == 2, R = eye(n) + A;    end
           if s > 2, R = R*(eye(n) + A); end
       end
    end

end  % end while

if refine && s >= 2
    switch rform
       % Two equivalent formulas, since R, Z0 both functions of A.
       % On test2, same errors except one matrix where "1" slightly better.
       case 1, R = Z0/R;
       case 2, R = R\Z0;
    end
else
    R = A - eye(n);
end
X = 2^s*logm_pf(R,m);
if isreal(A), X = real(X); end

cost = 4*cost*n^3 + 8*m*n^3/3;
if refine && s >= 2
    cost = cost + 2*(s-2/3)*n^3;
end

% Nested functions.
%%%%%%%%%%%%%%%%%%%%%%%%%
function S = logm_pf(A,m)
%LOGM_PF   Pade approximation to matrix log by partial fraction expansion.
%   LOGM_PF(A,m) is an [m/m] Pade approximant to LOG(EYE(SIZE(A))+A).

[nodes,wts] = gauss_legendre(m);
% Convert from [-1,1] to [0,1].
nodes = (nodes + 1)/2;
wts = wts/2;

S = zeros(n);
for j=1:m
    S = S + wts(j)*(A/(eye(n) + nodes(j)*A));
end
end

function [x,w] = gauss_legendre(n)
%GAUSS_LEGENDRE  Nodes and weights for Gauss-Legendre quadrature.
%   [X,W] = GAUSS_LEGENDRE(N) computes the nodes X and weights W
%   for N-point Gauss-Legendre quadrature.

% Reference:
% G. H. Golub and J. H. Welsch, Calculation of Gauss quadrature
% rules, Math. Comp., 23(106):221-230, 1969.

i = 1:n-1;
v = i./sqrt((2*i).^2-1);
[V,D] = eig( diag(v,-1)+diag(v,1) );
x = diag(D);
w = 2*(V(1,:)'.^2);
end

end
